"""
Command line interface for the simiotics_s3 tool
"""

import argparse

from simiotics.client import client_from_env
from simiotics.registry.data_pb2 import *

from . import data, sources

def parse_sources_list(args: argparse.Namespace) -> None:
    """
    Handler for the "sources list" command
    """
    simiotics_client = client_from_env()
    registered_sources = sources.list_data_sources(simiotics_client, args.offset, args.num_items)
    print('*** Sources ***')
    for i, source in enumerate(registered_sources):
        print('*** Source {} ***'.format(args.offset + i))
        print(source)

def parse_sources_get(args: argparse.Namespace) -> None:
    """
    Handler for the "sources get" command
    """
    simiotics_client = client_from_env()
    registered_source = sources.get_data_source(simiotics_client, args.id)
    print('*** Source ***')
    print(registered_source)

def parse_sources_update(args: argparse.Namespace) -> None:
    """
    Handler for the "sources update" command
    """
    simiotics_client = client_from_env()
    registered_source = sources.update_data_source(simiotics_client, args.id, args.message)
    print('*** Source ***')
    print(registered_source)

def parse_sources_create(args: argparse.Namespace) -> None:
    """
    Handler for the "sources create" command
    """
    simiotics_client = client_from_env()
    registered_source = sources.create_data_source(simiotics_client, args.id, args.s3_path)
    print('*** Source registered ***')
    print(registered_source)

def parse_data_describe(args: argparse.Namespace) -> None:
    """
    Handler for the "data describe" command
    """
    simiotics_client = client_from_env()
    data_descriptions = data.describe_data(simiotics_client, args.source, args.ids)
    print('*** Data descriptions ***')
    for i, description in enumerate(data_descriptions):
        print('*** Sample {} ***'.format(i))
        print(description)

def parse_data_download(args: argparse.Namespace) -> None:
    """
    Handler for the "data download" command
    """
    simiotics_client = client_from_env()
    paths = data.download_data(simiotics_client, args.source, args.ids, args.dir)
    for path in paths:
        print(path)

def parse_data_register(args: argparse.Namespace) -> None:
    """
    Handler for the "data register" command
    """
    simiotics_client = client_from_env()
    raw_tags = args.tags.split(',')
    tags = dict([tuple(raw_tag.split('=')) for raw_tag in raw_tags])
    responses = data.register_data(simiotics_client, args.source, args.files, tags)
    print('*** Data registration ***')
    for i, response in enumerate(responses):
        print('*** Sample {} ***'.format(i))
        print(response)

def generate_argument_parser() -> argparse.ArgumentParser:
    """
    Generates argument parser for the simiotics_s3 CLI

    Args: None

    Returns: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser('Define and interact with Simiotics S3 data registries')
    subparsers = parser.add_subparsers(
        title='Actions',
        help='Define sources, and add and retrieve data from those sources',
    )

    sources_subcommand = subparsers.add_parser('sources')
    sources_subparser = sources_subcommand.add_subparsers(
        title='Data sources',
        help='Register, list, signal updates, and get descriptions of sources of data'
    )

    sources_create = sources_subparser.add_parser('create', help='Create a source')
    sources_create.add_argument(
        '-i',
        '--id',
        required=True,
        help='An identifier for the source -- this must be unique over all registered sources',
    )
    sources_create.add_argument(
        '-p',
        '--s3-path',
        required=True,
        help='S3 path under which individual samples are stored',
    )
    sources_create.set_defaults(func=parse_sources_create)

    sources_get = sources_subparser.add_parser('get', help='Get a previously created source')
    sources_get.add_argument(
        '-i',
        '--id',
        required=True,
        help='An identifier for the source -- this must be unique over all registered sources',
    )
    sources_get.set_defaults(func=parse_sources_get)

    sources_update = sources_subparser.add_parser('update', help='Mark an update to a source')
    sources_update.add_argument(
        '-i',
        '--id',
        required=True,
        help='An identifier for the source -- this must be unique over all registered sources',
    )
    sources_update.add_argument(
        '-m',
        '--message',
        help='Update message -- specifies the nature of the update',
    )
    sources_update.set_defaults(func=parse_sources_update)

    sources_list = sources_subparser.add_parser('list', help='List previously created sources')
    sources_list.add_argument(
        '-o',
        '--offset',
        type=int,
        default=0,
        help='Offset from which listing should start',
    )
    sources_list.add_argument(
        '-n',
        '--num-items',
        type=int,
        default=10,
        help='Number of items to list',
    )
    sources_list.set_defaults(func=parse_sources_list)

    data_subcommand = subparsers.add_parser('data')
    data_subparser = data_subcommand.add_subparsers(
        title='Data samples',
        help='Register data against a Simiotics source and download datasets from that source'
    )

    data_describe = data_subparser.add_parser('describe', help='Describe the data in a data source')
    data_describe.add_argument(
        '-s',
        '--source',
        required=True,
        help='ID of the source'
    )
    data_describe.add_argument(
        '-i',
        '--ids',
        nargs='*',
        help='Optional IDs for data samples that you would like to restrict to in the description'
    )
    data_describe.set_defaults(func=parse_data_describe)

    data_download = data_subparser.add_parser('download', help='Describe the data in a data source')
    data_download.add_argument(
        '-s',
        '--source',
        required=True,
        help='ID of the source'
    )
    data_download.add_argument(
        '-i',
        '--ids',
        nargs='*',
        help='Optional IDs for data samples that you would like to restrict to in the description',
    )
    data_download.add_argument(
        '-d',
        '--dir',
        default='./',
        help='Directory into which data should be downloaded (defaults to current directory)',
    )
    data_download.set_defaults(func=parse_data_download)

    data_register = data_subparser.add_parser('register', help='Register data against a source')
    data_register.add_argument(
        '-s',
        '--source',
        required=True,
        help='ID of the source',
    )
    data_register.add_argument(
        '-t',
        '--tags',
        type=str,
        required=False,
        default='',
        help=(
            'Comma-separated list of <key>=<value> pairs specifying tags to be assocated to the '
            'images being registered (e.g. format=jpg,class=human)'
        )
    )
    data_register.add_argument(
        'files',
        nargs='+',
        help='Paths describing local files to be uploaded',
    )
    data_register.set_defaults(func=parse_data_register)

    return parser

def main() -> None:
    """
    Runs the simiotics_s3 CLI
    """
    parser = generate_argument_parser()
    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



"""
Source-related functionality for simiotics_s3 tool
"""

from typing import Any, Iterator, List, Optional

from simiotics.client import Simiotics
import simiotics.registry.data_pb2 as data_pb2

def list_data_sources(
        simiotics_client: Simiotics,
        offset: int,
        num_items: int
    ) -> List[data_pb2.Source]:
    """
    Lists source in a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    offset
        Offset from which list should start
    num_items
        Number of items to list

    Returns: Source object
    """
    request = data_pb2.ListSourcesRequest(
        version=simiotics_client.client_version,
        offset=offset,
        num_items=num_items,
    )
    response = simiotics_client.data_registry.ListSources(request)

    return response.sources

def update_data_source(
        simiotics_client: Simiotics,
        source_id: str,
        message: str,
    ) -> data_pb2.Source:
    """
    Gets a source from a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source you would like to register
    message
        Update message signifying the nature of the marked update

    Returns: Source object
    """
    request = data_pb2.UpdateSourceRequest(
        version=simiotics_client.client_version,
        id=source_id,
        notes=message,
    )
    response = simiotics_client.data_registry.UpdateSource(request)

    return response.source

def get_data_source(
        simiotics_client: Simiotics,
        source_id: str,
    ) -> data_pb2.Source:
    """
    Gets a source from a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source you would like to register

    Returns: Source object
    """
    request = data_pb2.GetSourceRequest(
        version=simiotics_client.client_version,
        id=source_id,
    )
    response = simiotics_client.data_registry.GetSource(request)

    return response.source

def create_data_source(
        simiotics_client: Simiotics,
        source_id: str,
        s3_root: str,
    ) -> data_pb2.Source:
    """
    Registers an S3 data source against a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source you would like to register
    s3_root
        Root under which all source samples may be found

    Returns: Source object
    """
    source = data_pb2.Source(
        id=source_id,
        source_type=data_pb2.Source.SourceType.SOURCE_S3,
        data_access_spec=s3_root,
    )

    request = data_pb2.RegisterSourceRequest(
        version=simiotics_client.client_version,
        source=source,
    )
    response = simiotics_client.data_registry.RegisterSource(request)

    return response.source



"""
Data-related functionality for simiotics_s3 tool
"""
import os
from typing import Iterator, List, Optional, Tuple
import uuid

import boto3
from simiotics.client import Simiotics
import simiotics.registry.data_pb2 as data_pb2

from . import sources

def describe_data(
        simiotics_client: Simiotics,
        source_id: str,
        data_ids: Optional[List[str]],
    ) -> Iterator[data_pb2.Datum]:
    """
    Describe data registered under a source in a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source under which the data is registered
    data_ids
        Optional IDs for the data samples you would like to restrict the description to

    Returns: Iterator over the data descriptions
    """
    request = data_pb2.GetDataRequest(
        version=simiotics_client.client_version,
        source_id=source_id,
    )
    if data_ids is not None:
        request.ids.extend(data_ids)

    responses = simiotics_client.data_registry.GetData(request)

    return (response.datum for response in responses)

def download_data(
        simiotics_client: Simiotics,
        source_id: str,
        data_ids: Optional[List[str]],
        target_dir: str,
    ) -> Iterator[str]:
    """
    Describe data registered under a source in a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source under which the data is registered
    data_ids
        Optional IDs for the data samples you would like to restrict the description to
    target_dir
        Directory into which data should be downloaded

    Returns: Iterator over the data descriptions
    """
    request = data_pb2.GetDataRequest(
        version=simiotics_client.client_version,
        source_id=source_id,
    )
    if data_ids is not None:
        request.ids.extend(data_ids)

    responses = simiotics_client.data_registry.GetData(request)

    s3_client = boto3.client('s3')

    paths = []
    for response in responses:
        datum = response.datum

        object_path = datum.content[len('s3://'):].split('/')
        bucket = object_path[0]
        key = '/'.join(object_path[1:])

        target_filename = object_path[-1]
        target_filepath = os.path.join(target_dir, target_filename)
        try:
            s3_client.download_file(bucket, key, target_filepath)
            paths.append(target_filepath)
        except Exception as err:
            print(str(err))

    return paths

def register_data(
        simiotics_client: Simiotics,
        source_id: str,
        files: List[str],
        tags: List[Tuple[str, str]]
    ) -> Iterator[data_pb2.RegisterDataResponse]:
    """
    Register files under a source in a Simiotics data registry

    Args:
    simiotics_client
        Simiotics client -- see the simiotics.client module
    source_id
        String identifying the source under which the data is registered
    files
        Paths to files that should be registered
    tags
        List of key-value ordered pairs representing the tags that should be associated to each
        of the files when registered against the data registry

    Returns: Iterator over the responses to the registration requests
    """
    source = sources.get_data_source(simiotics_client, source_id)
    source_root = source.data_access_spec
    s3_prefix = 's3://'
    if source_root[:len(s3_prefix)] == s3_prefix:
        source_root = source_root[len(s3_prefix):]
    split_root = source_root.split('/')
    bucket = split_root[0]
    root_from_bucket = ''
    if len(split_root) > 1:
        root_from_bucket = '/'.join(split_root[1:])

    s3_client = boto3.client('s3')

    def request_iterator():
        for data_file in files:
            datum_id = str(uuid.uuid4())
            key = os.path.join(root_from_bucket, datum_id)
            successful = False
            try:
                s3_client.upload_file(
                    data_file,
                    bucket,
                    key,
                    ExtraArgs={
                        'Metadata': {
                            'filename': data_file,
                        },
                    },
                )
                successful = True
            except Exception as err:
                print(str(err))

            if successful:
                tags['filename'] = data_file
                datum = data_pb2.Datum(
                    id=datum_id,
                    source=source,
                    content='s3://{}/{}'.format(bucket, key),
                    tags=tags,
                )
                request = data_pb2.RegisterDataRequest(
                    version=simiotics_client.client_version,
                    datum=datum,
                )
                yield request

    responses = simiotics_client.data_registry.RegisterData(request_iterator())

    return responses
